/**
 * iSES Remote Lab SDK (c) by J.Dvorak, P.Kuriscak, F.Lustig
 *
 * This file is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 *
 * Copy of license along with attribution guidelines is distributed together with this work in LICENSE.txt file.
 * License text is also available online on <http://creativecommons.org/licenses/by-nc-sa/4.0/>.
 */

var VL = {};

// "Properties"
VL.ReadOnly = false;
VL.Password = "";
VL.Timeout = 1200;
VL.CurrentTimeout = VL.Timeout;

VL.ErrorMessage = "";
VL.Timeleft = -1;

VL.Queue = [];
VL.PreconnectQueue = [];
VL.PollList = [];
VL.PollCount = 0;
VL.PollSentTimestamp = 0;
VL.PollRecieved = 0;
VL.PollInterval = 1e9;

VL.EventListeners = [];
VL.EventInvokeDepth = 0; // How many calls of InvokeEvent are currently on the stack.
VL.Storage = [];

VL.RequestLittleEndian = false;  // Requests are big endian
VL.ResponseLittleEndian = false; // Responses are also in big endian

// CONSTANTS
VL.PROTOCOL_VERSION = 100;

VL.COMMAND_EXIT = 1;
VL.COMMAND_LISTDEVICES = 2;
VL.COMMAND_GET_SOURCE_DEVICEHANDLE = 3;
VL.COMMAND_GET_DESTINATION_DEVICEHANDLE = 4;
VL.COMMAND_MEASUREDEVICEVALUE = 5;
VL.COMMAND_SETDEVICEVALUE = 19;
VL.COMMAND_QUERYQUEUEPOS = 7;
VL.COMMAND_READLOG = 8;
VL.COMMAND_GENERATEHISTORY = 9;
VL.COMMAND_DELAYFLUSH = 10;
VL.COMMAND_GETQUEUE = 11;
VL.COMMAND_READFASTLOG = 12;
VL.COMMAND_STARTEXPERIMENT = 13;
VL.COMMAND_STOPEXPERIMENT = 14;
VL.COMMAND_READEXPERIMENT = 15;
VL.COMMAND_LISTEXPERIMENTS = 16;
VL.COMMAND_BURST_EXPERIMENT = 17;
VL.COMMAND_IS_EXPERIMENT_RUNNING = 18;
VL.COMMAND_GET_DEVICEHANDLE = 20;

VL.LOGINCOMMAND_CLIENT = 1;
VL.LOGINCOMMAND_READONLYCLIENT = 3;

VL.FEATURE_READ = 1;
VL.FEATURE_WRITE = 2;

// Connection state constants
VL.CS  = {
	INIT : 0,
	CONNECTED_CONTROL : 1,
	CONNECTED_WAIT : 2,
	ERROR : 3
};

VL.STATE = VL.CS.INIT;

VL.Debug = function(msg)
{
    document.getElementById("debug").innerHTML += msg + " ";
};

VL.DebugObj = function(obj)
{
    document.getElementById("debug").innerHTML += JSON.stringify(obj, null, 4);
};

VL.Connect = function(url)
{
   // Apply whatever timeout was configured previously
   VL.CurrentTimeout = VL.Timeout;

  // 2 DO : modernizr support?
  if ("WebSocket" in window)
  {
	 if (VL.STATE != VL.CS.INIT)
	 {
		VLH.Error("ERR_INCONSISTENT");
		return;
	 }

     VL.ws = new WebSocket(url);
     VL.ws.binaryType = "arraybuffer";

     VL.ws.onopen = function()
     {
        // Connect command
        VL.RequestStart(5);
        VL.WriteInt16(VL.PROTOCOL_VERSION);
        VL.WriteInt8(VL.ReadOnly ? VL.LOGINCOMMAND_READONLYCLIENT : VL.LOGINCOMMAND_CLIENT);
        VL.WriteString(VL.Password);
        VL.RequestSendBlind();
     };

     VL.ws.onmessage = function (evt)
     {

        if (VL.STATE == VL.CS.INIT) {
                VL.ResponseStart(evt.data);
                if ((VL.ResponseBufferView.byteLength != 1) || (VL.ReadInt8() != 0)) {
					VLH.Error("ERR_REFUSED");
                    VL.ws.close();
                    return;
                }

				VLH.ChangeState(VL.CS.CONNECTED_WAIT);

				VL.PollControl();
				setInterval(VL.DisconnectIfInactive,1000);
			    VL.RegisterEventListener("USER_ACTIVITY",function() {VL.CurrentTimeout = VL.Timeout;},this);

                //Process delayed request queue
                while (VL.PreconnectQueue.length > 0) {
                    var request = VL.PreconnectQueue.shift();
                    VL.RequestBuffer = request.data;
                    VL.RequestSend(request.responseCallback, request.callback);
                }
        }
        else {
            VL.ResponseStart(evt.data);
            var response = VL.Queue.shift();

            if (typeof(response.responseCallback)=="function")
            {
                var result = response.responseCallback();
                if (typeof(response.callback)=="function")
                {
                    response.callback(result);
                }
            }
            else
            {
                if (typeof(response.callback)=="function")
                {
                    response.callback();
                }
            }
        }
     };

     VL.ws.onclose = function()
     {
		VLH.Error("ERR_DISCONNECT");
     };

     VL.ws.onerror = function(evt)
     {
		 VLH.Error("#"+evt.reason);
     }
  }
  else
  {
	  VLH.Error("ERR_INCOMPATIBLE");
  }
};

/* EVENTS */

VL.RegisterEventListener = function(eventName,listener,object)
{
	if (eventName instanceof Array) {
		var eArray = [];
		for (var i = 0; i < eventName.length; ++i) {
			var e = VL.RegisterEventListener(eventName[i], listener, object);
			eArray.push(e);
		}
		return eArray;
	}

	if (eventName!=""){
		var e = {};
		e.eventName = eventName;
		e.listener = listener;
		e.object = object;
		VL.EventListeners.push(e);
		return e;
	} else {
		return {};
	}
};

VL.ClearEventListener = function(regInfo)
{
	if (regInfo instanceof Array) {
		for (var i = 0; i < regInfo.length; ++i){
			regInfo[i].eventName = "";
		}
	} else {
		regInfo.eventName = "";
	}
	VL.PruneEventListeners();
};

/* If possible, removes all event listeners whose names were cleared */
VL.PruneEventListeners = function()
{
	// We can not reorder the array while it is processed by InvokeEvents
	if (VL.EventInvokeDepth > 0) {
		return;
	}
	for (var i = 0; i < VL.EventListeners.length; i++){
		if (VL.EventListeners[i].eventName==""){
			VL.EventListeners.splice(i,1);
			i--;
		}
	}
};


VL.InvokeEvent = function(eventName)
{
	if (eventName instanceof Array) {
		for (var i = 0; i < eventName.length; ++i) {
			VL.InvokeEvent(eventName[i]);
		}
		return;
	}

	if (eventName!=""){
		++VL.EventInvokeDepth;
		for (var i = 0; i < VL.EventListeners.length; i++){
			if (VL.EventListeners[i].eventName==eventName){
				$.proxy(VL.EventListeners[i].listener,VL.EventListeners[i].object)(eventName);
			}
		}
		--VL.EventInvokeDepth;

		// Remove listeners which were cleared from within the callback.
		VL.PruneEventListeners();
	}
};

/* POLL COMMANDS */
VL.Poll = function(handle,interval,callback)
{
    var newPolling = VL.PollList.length == 0;
    var found = false;
    VL.PollInterval = 1e9;

    for (var i = 0 ; i < VL.PollList.length ; i++)
    {
        if (VL.PollList[i].Handle == handle)
        {
            VL.PollList[i].Interval = Math.min(VL.PollList[i].Interval,interval);
            VL.PollList[i].Callbacks.push(callback);
            found = true;
        }
    }

    if (!found)
    {
        var newPoll = {};
        newPoll.Handle = handle;
        newPoll.Interval = interval;
        newPoll.Callbacks = [];
        newPoll.Callbacks.push(callback);
        VL.PollList.push(newPoll);
    }

    for (i = 0 ; i < VL.PollList.length ; i++ )
    {
        VL.PollInterval = Math.min(VL.PollInterval,VL.PollList[i].Interval);
    }

    if (newPolling)
    {
        VL.PollSend();
    }
};

VL.UnregisterPoll = function(handle,callback){
	for (var i = 0 ; i < VL.PollList.length ; i++)
		{
			if (VL.PollList[i].Handle == handle)
			{
				for (var j = 0 ; j < VL.PollList[i].Callbacks.length ; j++){
					if (VL.PollList[i].Callbacks[j]==callback){
						VL.PollList[i].Callbacks.splice(j, 1);
						return;
					}
				}
			}
		}
};

VL.PollSend = function()
{
    if((VL.STATE == VL.CS.CONNECTED_CONTROL) || (VL.STATE == VL.CS.CONNECTED_WAIT))
    {
        VL.PollCount = VL.PollList.length;
        VL.PollRecieved = 0;
        VL.PollSentTimestamp = new Date().getTime();

        VL.RequestStart(3*VL.PollCount);

        for ( var i = 0 ; i < VL.PollCount ; i++ ){
            VL.WriteInt8(VL.COMMAND_MEASUREDEVICEVALUE);
            VL.WriteInt16(VL.PollList[i].Handle);
            var request = {};
            request.responseCallback = VL.MeasureDeviceValueResponse;
            request.callback = VL.PollCallback;
            VL.Queue.push(request);
        }

        VL.ws.send(VL.RequestBuffer);
    }
    else
    {
        setTimeout(VL.PollSend,500);
    }
};

VL.PollCallback = function(result)
{
    var itemReceived = VL.PollRecieved;

    for ( var i = 0 ; i < VL.PollList[itemReceived].Callbacks.length ; i++)
    {
        if(typeof(VL.PollList[itemReceived].Callbacks[i]=="function"))
        {
            VL.PollList[itemReceived].Callbacks[i](result);
        }
    }

    VL.PollRecieved++;
    if (VL.PollRecieved >= VL.PollCount)
    {
        var milisToGo = (new Date().getTime())-VL.PollSentTimestamp-VL.PollInterval;
        if (milisToGo>0)
        {
            VL.PollSend();
        }
        else
        {
            setTimeout(VL.PollSend,-milisToGo);
        }
    }
};

VL.PollControl = function()
{
	// Register itself for queue update even if we are read-only
	// to prevent the connection from being aborted because of
	// no data transfer.

	if((VL.STATE == VL.CS.CONNECTED_CONTROL) || (VL.STATE == VL.CS.CONNECTED_WAIT))
	{
		VL.QueryQueuePos(VL.PollControlCallback);
	}
};
VL.PollControlCallback = function(result)
{
	if (! VL.ReadOnly) {
		VL.Timeleft = result.Timeleft;

		if ((result.Position == 0) && (VL.STATE == VL.CS.CONNECTED_WAIT))
		{
			VLH.ChangeState(VL.CS.CONNECTED_CONTROL);
		}
		else
		{
			if ((VL.STATE == VL.CS.CONNECTED_CONTROL) && (result.Position!=0))
			{
				VLH.ChangeState(VL.CS.CONNECTED_WAIT);
			}
		}

		VL.InvokeEvent("QUEUE_UPDATED");
	}

	setTimeout(VL.PollControl,1000);
};

VL.DisconnectIfInactive = function()
{
	VL.CurrentTimeout--;
	if (VL.CurrentTimeout<0)
	{
		VL.Disconnect(function(){});
		VLH.Error("ERR_TIMEOUT");
	}
};

/* PROTOCOL COMMANDS */

/**
 *
 * @param callback
 * @constructor
 */
VL.Disconnect = function(callback)
{
    VL.RequestStart(1);
    VL.WriteInt8(VL.COMMAND_EXIT);
    VL.RequestSend(function () {},callback);
	VLH.Error("ERR_DISCONNECT");
};

/**
 * Test test
 * @param handle Handle
 * @param callback callback
 * @constructor
 */
VL.MeasureDeviceValue = function(handle,callback)
{
    VL.RequestStart(3);
    VL.WriteInt8(VL.COMMAND_MEASUREDEVICEVALUE);
    VL.WriteInt16(handle);
    VL.RequestSend(VL.MeasureDeviceValueResponse,callback);
};
VL.MeasureDeviceValueResponse = function()
{
    var result = {};
    result.Value = VL.ReadFloat32();
    result.Timestamp = VL.ReadInt32();
    return result;
};

VL.QueryQueuePos = function(callback)
{
    VL.RequestStart(1);
    VL.WriteInt8(VL.COMMAND_QUERYQUEUEPOS);
    VL.RequestSend(VL.QueryQueuePosResponse,callback);
};
VL.QueryQueuePosResponse = function()
{
    var result = {};
    result.Position = VL.ReadInt32();
    result.Timeleft = VL.ReadInt32();
    return result;
};


VL.SetDeviceValue = function(handle,value,callback)
{
    VL.RequestStart(7);
    VL.WriteInt8(VL.COMMAND_SETDEVICEVALUE);
    VL.WriteInt16(handle);
    VL.WriteFloat32(value);
    VL.RequestSend(VL.SetDeviceValueResponse,callback);
};

VL.SetDeviceValueResponse = function()
{
	var result = {};
	result.Timestamp = VL.ReadInt32();
	return result;
};


VL.StartExperiment = function(experimentId, experimentSelector, maxLength, userValue, cyclic, handles, callback)
{
    VL.RequestStart(20+handles.length*2);
    VL.WriteInt8(VL.COMMAND_STARTEXPERIMENT);
    VL.WriteInt32(experimentSelector);
    VL.WriteInt32(experimentId);
    VL.WriteInt32(maxLength);
    VL.WriteInt32(userValue);
    VL.WriteInt8( cyclic ? 1 : 0 );
    VL.WriteInt16(handles.length);
    for ( var q = 0 ; q < handles.length ; q ++ ){
        VL.WriteInt16(handles[q]);
    }
    VL.RequestSend(VL.StartExperimentResponse,callback);
};
VL.StartExperimentResponse = function()
{
    var result = {};
    result.Id =  VL.ReadInt32();
    return result;
};


VL.ListExperiments = function(experimentSelector,minNum,maxNum,callback)
{
    VL.RequestStart(13);
    VL.WriteInt8(VL.COMMAND_LISTEXPERIMENTS);
    VL.WriteInt32(experimentSelector);
    VL.WriteInt32(minNum);
    VL.WriteInt32(maxNum);
    VL.RequestSend(VL.ListExperimentsResponse,callback);
};
VL.ListExperimentsResponse = function()
{
    var result = {};
    var experimentCount = VL.ReadInt32();
    result.Experiments = new Array(experimentCount);
    for (var j = 0 ; j < experimentCount ; j++){
        var experiment = {};
        experiment.Id = VL.ReadInt32();
        experiment.Running = VL.ReadInt8() != 0;
        experiment.StartDateTimeInSeconds = VL.ReadInt32();
        experiment.EndDateTimeInSeconds = VL.ReadInt32();
        experiment.IP = new Array(4);
        experiment.IP[0] = VL.ReadInt8();
        experiment.IP[1] = VL.ReadInt8();
        experiment.IP[2] = VL.ReadInt8();
        experiment.IP[3] = VL.ReadInt8();
        result.Experiments[j] = experiment;
    }
    return result;
};


VL.ReadExperiment = function(experimentId,handles,callback)
{
    VL.RequestStart(9+handles.length*10);
    VL.WriteInt8(VL.COMMAND_READEXPERIMENT);
    VL.WriteInt32(experimentId);
    VL.WriteInt32(handles.length);
    for ( var q = 0 ; q < handles.length ; q++ ){
        VL.WriteInt16(handles[q].Handle);
        VL.WriteInt32(handles[q].StartIndex);
        VL.WriteInt32(handles[q].SampleLimit);
    }
    VL.RequestSend(function(){return VL.ReadExperimentResponse(handles);},callback);
};
VL.ReadExperimentResponse = function(handles)
{
    var result = {};
    var deviceCount = VL.ReadInt32();
	result.ExperimentFound = deviceCount!=-1;
    if (deviceCount == -1)
    {
		result.DeviceData = new Array(handles.length);
		for (var j = 0 ; j < handles.length; j++){
			result.DeviceData[j] = {};
			result.DeviceData[j].Handle = handles[j];
			result.DeviceData[j].TotalLogSize = 0;
			result.DeviceData[j].Data = [];
		}
        return result;
    }
    else
    {
		if (deviceCount!=handles.length){
			VL.Disconnect(function(){});
			VLH.Error("ERR_INTERNAL");
			return null;
		}

        result.DeviceData = new Array(deviceCount);
        for ( var j = 0 ; j < deviceCount ; j++ ){
			var handle = VL.ReadInt16();
			var k = 0;
			for (; k < handles.length ; k++)
			{
				if (handles[k].Handle == handle) { break; }
			}

			if ((k==handles.length) || (result.DeviceData[k]!=undefined)) {
				VL.Disconnect(function(){});
				VLH.Error("ERR_INTERNAL");
				return null;
			}

            result.DeviceData[k] = {};
            result.DeviceData[k].Handle = handle;
            result.DeviceData[k].TotalLogSize = VL.ReadInt32();
            var recordCount = VL.ReadInt32();
            result.DeviceData[k].Data = new Array(recordCount);
            for ( var i = 0 ; i < recordCount ; i++ ){
                result.DeviceData[k].Data[i] = {};
                result.DeviceData[k].Data[i].Index = VL.ReadInt32();
                result.DeviceData[k].Data[i].Time = VL.ReadInt32() / 100.0;
                result.DeviceData[k].Data[i].Value = VL.ReadFloat32();
            }
        }
        return result;
    }
};


VL.StopExperiment = function(callback)
{
    VL.RequestStart(1);
    VL.WriteInt8(VL.COMMAND_STOPEXPERIMENT);
    VL.RequestSend(VL.StopExperimentResponse,callback);
};
VL.StopExperimentResponse = function()
{
    var result = {};
    result.Id = VL.ReadInt32();
    return result;
};


VL.ExperimentRunning = function(id,callback)
{
    VL.RequestStart(5);
    VL.WriteInt8(VL.COMMAND_IS_EXPERIMENT_RUNNING);
    VL.WriteInt32(id);
    VL.RequestSend(VL.ExperimentRunningResponse,callback);
};
VL.ExperimentRunningResponse = function()
{
    var result = {};
    result.isRunning = VL.ReadInt8() != 0;
    return result;
};


VL.QueryLog = function(handle,timeOrDuration,mode,modeParams,callback)
{
    VL.RequestStart(13);
    VL.WriteInt8(VL.COMMAND_READLOG);
    VL.WriteInt16(handle);
    VL.WriteInt32(timeOrDuration);
    VL.WriteInt16(mode);
    VL.WriteInt32(modeParams);
    VL.RequestSend(function() {return VL.QueryLogResponse(handle);},callback);
};
VL.QueryLogResponse = function(handle)
{
    var result = {};
    var count = VL.ReadInt32();
    result.DeviceData = new Array(1);
	result.DeviceData[0] = {};
	result.DeviceData[0].Handle = handle;
	result.DeviceData[0].TotalLogSize = count;
	result.DeviceData[0].Data = new Array(count);

    for ( var i = 0 ; i < count ; i++ ){
        result.DeviceData[0].Data[i] = {};
		result.DeviceData[0].Data[i].Index = i;
		result.DeviceData[0].Data[i].Time = VL.ReadInt32()*1000;
		result.DeviceData[0].Data[i].Value = VL.ReadFloat32();
    }
    return result;
};


VL.QueryFastLog = function(handle, mode, lengthOrIndex, limit ,callback)
{
    VL.RequestStart(10);
    VL.WriteInt8(VL.COMMAND_READFASTLOG);
    VL.WriteInt16(handle);
    VL.WriteInt8(mode);
    VL.WriteInt32(lengthOrIndex);
    VL.WriteInt16(limit);
    VL.RequestSend(VL.QueryFastLogResponse,callback);
};
VL.QueryFastLogResponse = function()
{
    var result = {};
    var count = VL.ReadInt32();
    result.Data = new Array(count);
    for ( var x = 0 ; x < count ; x++ ){
        result.Data[count-x-1] = {};
        result.Data[count-x-1].IndexOrMilliseconds = VL.ReadInt32Unsigned();
        result.Data[count-x-1].Value = VL.ReadFloat32();
    }

    if (count>0){
        result.LastIndex = VL.ReadInt32Unsigned();
    }
    return result;
};

VL.QueryQueueList = function(callback)
{
    VL.RequestStart(1);
    VL.WriteInt8(VL.COMMAND_GETQUEUE);
    VL.RequestSend(VL.QueryQueueListResponse,callback);
};
VL.QueryQueueListResponse = function()
{
    var result = {};
    var count = VL.ReadInt32();
    result.Users = new Array(count);
    for ( var i = 0 ; i < count ; i++ ){
        result.Users[i] = {};
        result.Users[i].IP = new Array(4);
        result.Users[i].IP[0] = VL.ReadInt8();
        result.Users[i].IP[1] = VL.ReadInt8();
        result.Users[i].IP[2] = VL.ReadInt8();
        result.Users[i].IP[3] = VL.ReadInt8();
    }
    return result;
};


VL.BurstExperiment = function(name,duration,callback)
{
    VL.RequestStart(5+name.length+2);
    VL.WriteInt8(VL.COMMAND_BURST_EXPERIMENT);
    VL.WriteString(name);
    VL.WriteInt32(duration);
    VL.RequestSend(VL.BurstExperimentResponse,callback);
};
VL.BurstExperimentResponse = function()
{
    var result = {};
    result.Id = VL.ReadInt32();
    return result;
};


VL.ListDevices = function(callback)
{
    VL.RequestStart(1);
    VL.WriteInt8(VL.COMMAND_LISTDEVICES);
    VL.RequestSend(VL.ListDevicesResponse,callback);
};
VL.ListDevicesResponse = function()
{
    var result = {};
    var count = VL.ReadInt32();
    result.Devices = new Array(count);
    for ( var i = 0 ; i < count ; i++ )
    {
        var features = VL.ReadInt8();
        result.Devices[i] = {};
        result.Devices[i].SupportsRead = (features & VL.FEATURE_READ) == VL.FEATURE_READ;
        result.Devices[i].SupportsWrite = (features & VL.FEATURE_WRITE) == VL.FEATURE_WRITE;
        result.Devices[i].DeviceName = VL.ReadString();
        result.Devices[i].DeviceIdentity = VL.ReadString();
    }
    return result;
};


VL.GetSourceDeviceHandle = function(pinName,callback)
{
    VL.RequestStart(3+pinName.length);
    VL.WriteInt8(VL.COMMAND_GET_SOURCE_DEVICEHANDLE);
    VL.WriteString(pinName);
    VL.RequestSend(VL.GetSourceDeviceHandleResponse,callback);
};
VL.GetSourceDeviceHandleResponse = function()
{
    var result = {};
    result.Handle = VL.ReadInt16();
    return result;
};

VL.GetDestinationDeviceHandle = function(pinName,callback)
{
    VL.RequestStart(3+pinName.length);
    VL.WriteInt8(VL.COMMAND_GET_DESTINATION_DEVICEHANDLE);
    VL.WriteString(pinName);
    VL.RequestSend(VL.GetDestinationDeviceHandleResponse,callback);
};
VL.GetDestinationDeviceHandleResponse = function()
{
    var result = {};
    result.Handle = VL.ReadInt16();
    return result;
};

VL.GetDeviceHandle = function(pinName,callback)
{
	VL.RequestStart(3+pinName.length);
	VL.WriteInt8(VL.COMMAND_GET_DEVICEHANDLE);
	VL.WriteString(pinName);
	VL.RequestSend(VL.GetDeviceHandleResponse,callback);
};
VL.GetDeviceHandleResponse = function()
{
	var result = {};
	result.Handle = VL.ReadInt16();
	return result;
};

/* LOW LEVEL I/O COMMANDS */

VL.RequestStart = function(len)
{
    VL.RequestBuffer = new ArrayBuffer(len);
    VL.RequestBufferView = new DataView(VL.RequestBuffer);
    VL.RequestBufferPos = 0;
};

VL.RequestSendBlind = function()
{
    VL.ws.send(VL.RequestBuffer);
};

VL.RequestSend = function(responseCallback,callback)
{
    var request = {};
    if (VL.STATE == VL.CS.INIT)
    {
        // If not connected, temporarily store the requests to our queue,
        // They will be processed in onopen event
        request.data = VL.RequestBuffer;
        request.responseCallback = responseCallback;
        request.callback = callback;
        VL.PreconnectQueue.push(request);
    }
    else
    {
		if ((VL.STATE == VL.CS.CONNECTED_WAIT) || (VL.STATE == VL.CS.CONNECTED_CONTROL))
		{
        	VL.RequestTrueSend(VL.RequestBuffer,responseCallback,callback);
		}
    }
};

VL.RequestTrueSend = function(data,responseCallback,callback)
{
    VL.ws.send(VL.RequestBuffer);
    var request = {};
    request.responseCallback = responseCallback;
    request.callback = callback;
    VL.Queue.push(request);
};

VL.ResponseStart = function(data)
{
   VL.ResponseBufferView = new DataView(data);
   VL.ResponseBufferPos = 0;
};

VL.WriteInt8 = function(value)
{
    VL.RequestBufferView.setUint8(VL.RequestBufferPos,value);
    VL.RequestBufferPos += 1;
};

VL.WriteInt16 = function(value)
{
    VL.RequestBufferView.setUint16(VL.RequestBufferPos,value,VL.RequestLittleEndian);
    VL.RequestBufferPos += 2;
};

VL.WriteInt32 = function(value)
{
    VL.RequestBufferView.setInt32(VL.RequestBufferPos,value,VL.RequestLittleEndian);
    VL.RequestBufferPos += 4;
};

VL.WriteFloat32 = function(value)
{
    VL.RequestBufferView.setFloat32(VL.RequestBufferPos,value,VL.RequestLittleEndian);
    VL.RequestBufferPos += 4;
};

VL.WriteString = function(value)
{
    VL.WriteInt16(value.length);
    for (var i = 0; i < value.length ; i++ ){
        VL.WriteInt8(value.charCodeAt(i));
    }
};

VL.ReadInt8 = function()
{
    var result = VL.ResponseBufferView.getUint8(VL.ResponseBufferPos);
    VL.ResponseBufferPos += 1;
    return result;
};

VL.ReadInt16 = function()
{
    var result = VL.ResponseBufferView.getUint16(VL.ResponseBufferPos,VL.ResponseLittleEndian);
    VL.ResponseBufferPos += 2;
    return result;
};

VL.ReadInt32 = function()
{
    var result = VL.ResponseBufferView.getInt32(VL.ResponseBufferPos,VL.ResponseLittleEndian);
    VL.ResponseBufferPos += 4;
    return result;
};

VL.ReadInt32Unsigned = function()
{
    var result = VL.ResponseBufferView.getUint32(VL.ResponseBufferPos,VL.ResponseLittleEndian);
    VL.ResponseBufferPos += 4;
    return result;
};

VL.ReadFloat32 = function()
{
    var result = VL.ResponseBufferView.getFloat32(VL.ResponseBufferPos,VL.ResponseLittleEndian);
    VL.ResponseBufferPos += 4;
    return result;
};

VL.ReadString = function()
{
    var str = "";
    var length = VL.ReadInt16();
    for ( var i = 0 ; i < length ; i++ ){
        str += String.fromCharCode(VL.ReadInt8());
    }
    return str;
};

