/**
 * iSES Remote Lab SDK (c) by J.Dvorak, P.Kuriscak, F.Lustig
 *
 * This file is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 *
 * Copy of license along with attribution guidelines is distributed together with this work in LICENSE.txt file.
 * License text is also available online on <http://creativecommons.org/licenses/by-nc-sa/4.0/>.
 */

/* HELPERS */

var VLH = {};

VLH.CreateTag = function(tag,element,name,content,className){
contentText = content ? content : "";
	return $("<"+tag+">" ,{
		id : element.attr('id')+"_"+name,
		text : contentText,
		class : className
	}).appendTo(element);
};

VLH.ChangeState = function(state){
	VL.STATE = state;
	VL.InvokeEvent("STATE_CHANGE");
};

VLH.Error = function(message){
	VL.ErrorMessage = message;
	VLH.ChangeState(VL.CS.ERROR);
};

VLH.CreateCheckbox = function(element,name,label,checked){
return $("<input>", {
		id : element.attr('id')+"_"+name,
		type : "checkbox",
		value : 1,
		checked : checked
		}).appendTo($("<label>",{
					text : label+" "
					}).appendTo(element));
};

VLH.CreateOption = function(element,value,text,selected){
	if (selected) {
		return $("<option>",{
			value : value,
			text : text,
			selected : "selected",
			}).appendTo(element);
	} else {
		return $("<option>",{
			value : value,
			text : text
			}).appendTo(element);
	}
};

VLH.ScaleDeviceValue = function(value,device){
	if (VLH.IsScalable(device)){
		return VLH.ScaleValue(value, device.sMin, device.sMax, device.dMin, device.dMax);
	} else {
		return value;
	}
};

VLH.Clamp = function(value,min,max){
	return Math.max(Math.min(value,max),min);
};

VLH.IsScalable = function(device){
	return ((device.sMin!=null) && (device.sMax!=null) && (device.dMin!=null) && (device.dMax!=null))
};

VLH.ScaleValue = function (value, sMin, sMax, dMin, dMax) {
	if (value < sMin) {
		value = sMin;
	}

	if (value > sMax) {
		value = sMax;
	}

	var sRange = sMax - sMin;
	var dRange = dMax - dMin;

	if (sRange <= 0.000001) {
		return dMin;
	}

	var factor = (value - sMin) / sRange;
	return dMin + dRange * factor;
};

VLH.CenterAndSetRange = function(min,max,center,minRange,maxRange)
{
	if (center!== undefined){
		var centerDistance = Math.max(Math.abs(center-min),Math.abs(center-max));
		max = center+centerDistance;
		min = center-centerDistance;
	}

	var range = max-min;
	var realCenter = 0.5*(max+min);

	var halfRange = range*0.5;
	if (minRange!==undefined){
		halfRange = Math.max(halfRange,minRange*0.5);
	}

	if (maxRange !== undefined){
		halfRange = Math.min(halfRange,maxRange*0.5);
	}

	max = realCenter + halfRange;
	min = realCenter - halfRange;

	var result = {};
	result.min = min;
	result.max = max;
	return result;
};

VLH.InitDataFromDevices = function(devices)
{
	var colNames = [];
	for (var i = 0; i < devices.length ; i++){
		colNames.push(devices[i].deviceName+".time");
		colNames.push(devices[i].deviceName);
	}
	return VLH.InitData(colNames);
};

VLH.InitData = function(colNames){
	var data = {};
	data.cols = [];
	data.series = {};

	for (var i = 0; i < colNames.length; i++){
		if (colNames[i] in data.series) {
			alert("Loading data error : Column " + colNames[i] + " was specified more than once.")
		}
		var column = {};
		column.name = colNames[i];
		column.values = [];
		column.log = [];
		data.cols.push(column);
		data.series[colNames[i]] = column;
	}
	return data;
};

/**
 * Returns specified parameter if it is present in the structure and
 * default value otherwise.
 */
VLH.getParamWithDefault = function(params, name, defaultValue) {
	if (name in params) {
		return params[name];
	}
	else {
		return defaultValue;
	}
}

/**
 * Returns series with specified name from the data or shows alert if the
 * series does not exist.
 */
VLH.getSeries = function(data, name) {
	if (! (name in data.series)) {
		alert("Series " + name + " not found");
		return null;
	}
	return data.series[name];
};

/**
 * Calculates average of specified range of inValues and pushes the value to outValues
 * array and outRMSQ array (if not null)
 */
VLH.averageRange = function(inValues, startIndex, count, outValues, outRMSQ) {

	// Calculate the average value of the block.

	var afterEndIndex = startIndex + count;
	var sum = 0;

	for (var i = startIndex; i < afterEndIndex; ++i) {
		sum += inValues[i];
	}

	var average = sum / count;
	outValues.push(average);

	// If RMSQ is not wanted, we are done.

	if (outRMSQ === null) {
		return;
	}

	// For single sample there is no error.

	if (count == 1) {
		outRMSQ.push(0);
		return;
	}

	var error = 0;
	for (var i = startIndex; i < afterEndIndex; ++i) {
		var delta = inValues[i] - average;
		error += (delta * delta);
	}

	var rmsq = Math.sqrt(error / (count * (count - 1)));
	outRMSQ.push(rmsq);
};

VLH.GetDataCols = function(bigData,n){
	var data = {};
	data.cols  = [bigData.cols[2*n],bigData.cols[2*n+1]];
	return data;
};

VLH.LoadData = function(devices,data,state,callback){
	if ($.isEmptyObject(state)){
		VLH.setDeviceDefaults(devices);
		state.runID = 0;
	} else {
		state.runID++;
	}

	var waitForAll = true;
	for (var i = 0; i < devices.length ; i ++){
		if ((devices[i].loadMethod=='Fast') || (devices[i].loadMethod=='Poll')){
			waitForAll = false;
			break;
		}
	}

	var realCallback = callback;
	if (waitForAll){
		state.numDevicesLoaded = 0;
		state.numDevicesToLoad = devices.length;
		realCallback = function() {
			state.numDevicesLoaded++;
			if (state.numDevicesLoaded==state.numDevicesToLoad){
				callback();
			}
		};
	}

	(function(runID){
		VLH.GetDeviceHandles(devices, function(){
		if(state.runID!=runID) { return; }

		for (var i = 0; i < devices.length ; i++)
		{
			var dataCols = VLH.GetDataCols(data,i);
			switch(devices[i].loadMethod)
			{
				case 'Fast':
						VLH.checkOption(devices[i],"refresh",100);
						VLH.checkOption(devices[i],"timeLimit",10);
						VLH.LoadDataFast(devices[i],dataCols,state,runID,realCallback,true);
					break;
				case 'Poll':
						VLH.checkOption(devices[i],"refresh",100);
						VLH.checkOption(devices[i],"timeLimit",10);
						VLH.LoadDataPoll(devices[i],dataCols,state,runID,realCallback);
					break;
				case 'Experiment':
						if (typeof devices[i].slot == 'undefined'){
							alert("Loading data error : Device " + devices[i].deviceName + " loads Experiment, but has undefined slot.");
							return;
						}
						VLH.checkOption(devices[i],"sampleLimit",50000);
						VLH.LoadDataExperiment(devices[i],dataCols,state,runID,realCallback);
					break;
				case 'History':
						if ((typeof devices[i].beginTime == 'undefined') || (typeof devices[i].endTime == 'undefined')){
							alert("Loading data error : No begin or end time specified for device " + devices[i].deviceName);
							return;
						}

						VLH.checkOption(devices[i],"hourOfDay",-1);

						VLH.checkOption(devices[i],"steps",[
							[ 24 * 60 * 60 , "minute" ],
							[ 24 * 60 * 60 * 3 , "ten_minutes"],
							[ 24 * 60 * 60 * 20 , "hour" ],
							[ 24 * 60 * 60 * 60 , "three_hours" ],
							[ 24 * 60 * 60 * 120 , "six_hours" ]
						]);

						VLH.checkOption(devices[i],"timeLimit",16000000);
						if (devices[i].timeLimit > 16000000){
							alert("Cannot set timeLimit higher than 16000000 for device "+devices[i].deviceName);
							return;
						}

						VLH.LoadDataHistory(devices[i],dataCols,state,runID,realCallback);
					break;
				case undefined:
						alert("Loading data error : Device " + devices[i].deviceName + " has undefined loadMethod");
					break;
				default:
					alert("Loading data error : Device " + devices[i].deviceName + " has invalid loadMethod")
			}
		}
		});
	})(state.runID);
};


VLH.LoadDataHistory = function(device,data,state,runID,callback){
	var hourOfDay = VLH.ResolveValue(device.hourOfDay);
	var beginTime = VLH.ResolveValue(device.beginTime);
	var endTime = VLH.ResolveValue(device.endTime);

	beginTime = VLH.ParseDayOffset(beginTime);
	endTime = VLH.ParseDayOffset(endTime);

	var timeDifference = endTime - beginTime;

	if (timeDifference <= 0){
		data.cols[0].log.push(VLH.L("ERR_DATA_INTERVAL"));
		data.cols[1].log.push(VLH.L("ERR_DATA_INTERVAL"));
		callback();
		return;
	}

	if (timeDifference > device.timeLimit){
		data.cols[0].log.push(VLH.L("ERR_DATA_LENGTH"));
		data.cols[1].log.push(VLH.L("ERR_DATA_LENGTH"));
		callback();
		return;
	}

	// Prevent overflow even if the parameters are wrong.

	if (timeDifference > 0xffffff) {
		timeDifference = 0xffffff;
	}

	var mode;

	if (hourOfDay==-1){
		var maxStepInterval = -10000;
		var maxStep = "everything";

		for (var i = 0 ; i < device.steps.length ; i++ ){
			if((timeDifference>=device.steps[i][0]) && (maxStepInterval<device.steps[i][0])){
				maxStepInterval = device.steps[i][0];
				maxStep = device.steps[i][1];
			}
		}

		switch(maxStep){
			case "everything":
				mode = 24;
				break;
			case "minute":
				mode = 25;
				break;
			case "ten_minutes":
				mode = 26;
				break;
			case "hour":
				mode = 27;
				break;
			case "three_hours":
				mode = 28;
				break;
			case "six_hours":
				mode = 29;
				break;
			default:
				alert("Loading data error: Invalid step '"+maxStep+"' in device "+device.deviceName+". Please use one of 'everything', 'minute','ten_minutes','hour','three_hours' or 'six_hours'.");
				return;
		}
	} else {
		if ((hourOfDay>=0) && (hourOfDay<=23)){
			mode = hourOfDay;
		} else {
			alert("Loading data error: Invalid hourOfDay set in device "+device.deviceName+". Please use number between 0-23.");
			return;
		}
	}

	var modeParam = (timeDifference & 0xffffff) | ((mode & 31) << 24);

	VL.QueryLog(device.handle,beginTime,2,modeParam,function(result){
		if (state.runID!=runID){ return; }

		// If there are no data at the start of the interval, the server
		// will append additional data at the end to maintain the length.
		// While this might be usefull for some cases, we want a strict
		// subset of the interval.

		var dataLength = result.DeviceData[0].Data.length;
		while ((dataLength > 0) && (result.DeviceData[0].Data[dataLength - 1].Time / 1000) > endTime) {
			--dataLength;
		}

		data.cols[0].values = new Array(dataLength);
		data.cols[1].values = new Array(dataLength);

		for (var j = 0; j < dataLength; j++){
			data.cols[0].values[j] = result.DeviceData[0].Data[j].Time/1000.0;
			data.cols[1].values[j] = VLH.ScaleDeviceValue(result.DeviceData[0].Data[j].Value,device);
		}

		callback();
	});
};

VLH.ParseDayOffset = function(input){
	if (typeof input === "number") return input;
	if (input[input.length-1]=="d") {
		var date = new Date();
		var offset = input.substring(0,input.length-1);
		if (!isNaN(offset)) {

			return date.setDate(date.getDate() + parseInt(offset))/1000;
		} else {
			alert("Could not convert " + input + " to date.");
			return date.getTime()/1000;
		}
	} else {
		alert("Could not convert " + input + " to date. Maybe you are missing the 'd' suffix?");
		return new Date().getTime()/1000;
	}
};

VLH.LoadDataExperiment = function(device,data,state,runID,callback){
	if (!VL.Storage[device.slot]){
		callback();
		return;
	}

	var handles = [];
	var handle = {};
	handle.Handle = device.handle;
	handle.StartIndex = 0;
	handle.SampleLimit = device.sampleLimit;
	handles.push(handle);

	VL.ReadExperiment(VL.Storage[device.slot],handles, $.proxy(function(result){
		if (state.runID!=runID){ return; }
		if (result==null) { callback(); return; }
		if (!result.ExperimentFound){ callback(); return; }

		data.cols[0].values = new Array(result.DeviceData[0].Data.length);
		data.cols[1].values = new Array(result.DeviceData[0].Data.length);

		for (var j = 0; j < result.DeviceData[0].Data.length ; j++){
			data.cols[0].values[j] = result.DeviceData[0].Data[j].Time/1000.0;
			data.cols[1].values[j] = VLH.ScaleDeviceValue(result.DeviceData[0].Data[j].Value,device);
		}

		callback();
	},this));
};

// 2DO, possibly add "independent" polling
VLH.LoadDataPoll = function(device,data,state,runID,callback){

	var callbackInfo = {};
	callbackInfo.callback = function(result) {
		if(state.runID!=runID) {
			VL.UnregisterPoll(device.handle,callbackInfo.callback);
			return;
		}

		data.cols[0].values.push(result.Timestamp);
		data.cols[1].values.push(VLH.ScaleDeviceValue(result.Value,device));

		VLH.TrimData(data,device.timeLimit);
		callback();
	};

	VL.Poll(device.handle,device.refresh,callbackInfo.callback);
};

VLH.LoadDataFast = function(device,data,state,runID,callback,firstRun){
	device.lastDataTime = new Date().getTime();

	var queryData = function()
	{
		if(state.runID!=runID) {return;}
		VL.QueryFastLog(device.handle, 1 | 2 , device.lastIndex + 1, 0, function(result){
				if(state.runID!=runID) {return;}

				for (var j = 0; j < result.Data.length; j++) {
					var recalcValue = VLH.ScaleDeviceValue(result.Data[j].Value,device);
					data.cols[0].values.push(result.Data[j].IndexOrMilliseconds/1000.0);
					data.cols[1].values.push(recalcValue);
				}

				if (result.Data.length > 0) {
					device.lastIndex = result.LastIndex;
				}

				VLH.TrimData(data,device.timeLimit);

				var milisToGo = (new Date().getTime()) - device.lastDataTime - device.refresh;
				if (milisToGo > 0) {
					VLH.LoadDataFast(device,data,state,runID,callback,false);
				}
				else {
					setTimeout(function () {
						VLH.LoadDataFast(device,data,state,runID,callback,false);
					}, -milisToGo);
				}
				callback();
		});
	};

	if(firstRun) {
		VLH.GetDeviceIndices([device],queryData);
	}
	else {
		queryData();
	}
};

VLH.StopLoadingData = function(state){
	if (!$.isEmptyObject(state)){
		state.runID++;
	}
};

VLH.TrimData = function(data,timeLimit){
	var lastIndex = data.cols[0].values.length - 1;
		if (lastIndex >= 0) {
			var lastTime = data.cols[0].values[lastIndex];
			while ((data.cols[0].values.length > 0) && (data.cols[0].values[0] < (lastTime - timeLimit))) {
				data.cols[0].values.shift();
				data.cols[1].values.shift();
			}
		}
};


VLH.InitPostProc = function(postProcs){
	var state = [];
	for (var i = 0; i < postProcs.length;i++){
		state.push({});
	}
return state;
};

VLH.UpdateAverage = function(newValue,valueArray,length){
	valueArray.push(newValue);
	if (valueArray.length>length) { valueArray.shift(); }
	var sum = 0;
	for (var i = 0; i < valueArray.length;i++){
		sum += valueArray[i];
	}
	return sum/length;
};

VLH.PostProc = function(postProcs,data,state){
	for (var i = 0; i < postProcs.length;i++){
		var result = postProcs[i].func(postProcs[i].params,data,state[i]);
		data = VLH.DataMerge(data,result);
	}
	return data;
};

VLH.DataMerge = function(data,toAdd)
{
	var newData = {};

	// Create clone of the old data set.

	newData.cols = data.cols.concat([]);
	newData.series = $.extend({},data.series);

	// Add new records replacing existing ones if found

	for (name in toAdd.series) {

		// Do we already have such column?

		var seriesData = toAdd.series[name];
		if (name in newData.series) {

			// Find corresponding record in the indexed data and
			// replace it.

			for (var i = 0; i < newData.cols.length; ++i) {
				if (newData.cols[i].name == name) {
					newData.cols[i] = seriesData;
					break;
				}
			}
		}
		else {
			newData.cols.push(seriesData)
		}

		// Update the name based lookup.

		newData.series[name] = seriesData;
	}
	return newData;
};


VLH.ForEachDevice = function(devices,funToCall,deviceCallback,completeCallback){
	var deviceLoadInfo = {
		numDevicesToLoad : devices.length,
		numDevicesLoaded : 0,
		devices : devices,
		deviceCallback : deviceCallback,
		completeCallback : completeCallback
	};

	for (var i = 0; i < devices.length; i++) {
		funToCall(devices[i],VLH.ForEachDeviceCallback(deviceLoadInfo,i));
	}
};

VLH.ForEachDeviceCallback = function(deviceLoadInfo,i){
	return function(result) {
		deviceLoadInfo.deviceCallback(result,deviceLoadInfo.devices[i]);
		deviceLoadInfo.numDevicesLoaded++;
		if (deviceLoadInfo.numDevicesLoaded==deviceLoadInfo.numDevicesToLoad){
			deviceLoadInfo.completeCallback(deviceLoadInfo.devices);
		}
	}
};

VLH.setDeviceDefaults = function (devices) {
	for (var i = 0; i < devices.length; i++) {
		VLH.checkOption(devices[i],'sMin',null);
		VLH.checkOption(devices[i],'sMax',null);
		VLH.checkOption(devices[i],'dMin',null);
		VLH.checkOption(devices[i],'dMax',null);
		VLH.checkOption(devices[i],'deviceName','ERROR: No device name');
		VLH.checkOption(devices[i],'seriesName',devices[i].deviceName);
	}
};

VLH.GetDeviceHandles = function (devices, completeCallback) {
	VLH.ForEachDevice(devices,function(device,deviceCallback) {
		var deviceName = VLH.ResolveValue(device.deviceName);
		VL.GetDeviceHandle(deviceName,deviceCallback);
	}, function (result,device) {
		if (result.Handle == 0){
			var deviceName = VLH.ResolveValue(device.deviceName);
			alert('Error : Cannot find device ' + deviceName);
		} else {
			device.handle = result.Handle;
		}
	}, completeCallback);
};

VLH.ResolveValue = function(valueOrSlot){
	if (isNaN(valueOrSlot)){
		return valueOrSlot.substring(0,5)=='slot:' ? VL.Storage[valueOrSlot.substring(5)] : valueOrSlot;
	} else {
		return valueOrSlot;
	}
};

VLH.GetDeviceIndices = function (devices, completeCallback) {
	VLH.ForEachDevice(devices,function(device,deviceCallback) {
		VL.QueryFastLog(device.handle,2,1,1,deviceCallback);
	}, function (result,device) {
		device.lastIndex = result.LastIndex;
		device.startTime = result.Data[0].IndexOrMilliseconds;
	}, completeCallback);
};

/**
 * @brief Returns length of array if provided value is array and
 * 1 otherwise.
 */
VLH.getArrayLength = function(value) {
	if (value instanceof Array) {
		return value.length;
	}
	return 1;
};

/**
 * Returns index-th value from array with handling of scalar values as 1-sized arrays
 * and out-of range index by returning specified default.
 */
VLH.getArrayEntryWithDefault = function(value_array, index, default_value) {

	// For scalars behave as if it was array with single element.

	if (! (value_array instanceof Array)) {
		return (index == 0) ? value_array : default_value;
	}

	// For arrays check the index.

	return (index < value_array.length) ? value_array[index] : default_value;
};

/**
 * @brief Converts non-array value into single-entry array and
 * returns array unchanged.
 */
VLH.toArray = function(value) {
	if (value instanceof Array) {
		return value;
	}
	return [value];
};

VLH.exportToElement = function(){
	return { pre: null , post: VLH.exportToElementPost }
};

VLH.exportToElementPost = function(content,params) {
	if (data==null) { params.element.html(VLH.L("NO_DATA")); return;}
	params.element.html(content);
};

VLH.exportToFile = function(){
	return { pre: null , post: VLH.exportToFilePost }
};

VLH.exportToFilePost = function(content,params){
	if (content == null) {
		alert(VLH.L("NO_DATA"));
		return;
	}

	if (!("contentType" in params)){
		params.contentType = "text/plain;charset=utf-8";
	}

	if (!("fileName" in params)){
		params.fileName = "data.txt";
	}

	var blob = new Blob(["\ufeff" + content], {type: params.contentType });
	saveAs(blob, params.fileName);
};

VLH.exportToWindow = function(){
	return { pre: VLH.exportToWindowPre , post: VLH.exportToWindowPost };
};

VLH.exportToWindowPre = function(params) {
	params.windowObject = open();
};

VLH.exportToWindowPost = function(content, params) {

	if ((! ("windowObject" in params)) || (params.windowObject == null)) {
			return;
	}

	params.windowObject.document.clear();
	if (content == null) {
		params.windowObject.document.write(VLH.L("NO_DATA"))
	}
	else {
		params.windowObject.document.write(content);
	}
	params.windowObject.document.close();
};

VLH.getTimeFormater = function(params) {
	if (params.timeMode == "seconds") {
		return function(value) { return $.formatNumber(value,{format: params.timeFormat,locale:VLH.L("FORMAT_LOCALE")}); };
	}
	else if ((params.timeMode == "UTC") || (params.timeMode == "local")) {
		var tmp_date = new Date();
		return function(value) { tmp_date.setTime(value * 1000); return Flotr.Date.format(tmp_date, params.timeFormat, params.timeMode); };
	}
	else {
		alert("Unknown timeMode '" + params.timeMode + "'. Expected one from: seconds, UTC, local");
		return function(value) { return value; };
	}
};

VLH.getValueFormater = function(format) {
	if ($.isFunction(format)) {
		return format;
	}
	else {
		return function (value) { return $.formatNumber(value, {format: format, locale:VLH.L("FORMAT_LOCALE")}); };
	}
};

VLH.formatUniversal = function(data,params){
	VLH.checkOption(params,"beforeAll","");
	VLH.checkOption(params,"afterAll","");
	VLH.checkOption(params,"afterHead","");
	VLH.checkOption(params,"delimiterHead","");
	VLH.checkOption(params,"afterData","");
	VLH.checkOption(params,"beforeDataLine","");
	VLH.checkOption(params,"afterDataLine","");
	VLH.checkOption(params,"delimiterDataLine","");


	VLH.checkOption(params,"timeHead","Time");
	VLH.checkOption(params,"columnHeads","Series");
	VLH.checkOption(params,"timeMode", "seconds");
	VLH.checkOption(params,"timeFormat","0.000");
	VLH.checkOption(params,"valueFormats","0.000");

	var time_formater = VLH.getTimeFormater(params);

	var content = params.beforeAll;

	if (params.columnHeads instanceof Array){
		if(params.columnHeads.length!=data.cols.length/2){
			alert("Experiment data error : columnHeads parameter must be either single string, or array with number of strings equal to number of devices.");
		}
	}
	else
	{
		var head = params.columnHeads;
		params.columnHeads = [];
		for (var i = 0; i < data.cols.length/2; i++){
			params.columnHeads.push(head+" #"+(i+1));
		}
	}

	if (params.valueFormats instanceof Array){
		if(params.valueFormats.length!=data.cols.length/2){
			alert("Experiment data error : valueFormats parameter must be either single string, or array with number of strings equal to number of devices.");
		}
	}
	else
	{
		var format = params.valueFormats;
		params.valueFormats = [];
		for (var i = 0; i < data.cols.length/2; i++){
			params.valueFormats.push(format);
		}
	}

	// Prepare value formaters.

	var formaters = new Array();
	for (var i = 0; i < data.cols.length/2; ++i) {
		formaters.push(VLH.getValueFormater(params.valueFormats[i]));
	}

	// Detect if there was any error in any column. In that case
	// we will use the unsynchronized display and show error messages
	// instead of values in relevant columns.

	var was_error = false;
	for (var i = 0; i < data.cols.length; i++){
		was_error = was_error || (data.cols[i].log.length > 0);
	}

	if (VLH.checkSynchronousData(data) && (! was_error))
	{
		content += params.timeHead;
		for(var j = 0; j < data.cols.length/2; j++)
		{
			content += params.delimiterHead + params.columnHeads[j];
		}

		content += params.afterHead;

		for (var k = 0; k < data.cols[0].values.length;k++)
		{
			content += params.beforeDataLine;

			content += time_formater(data.cols[0].values[k]);
			for (var l = 0; l < data.cols.length;l += 2)
			{
				content += params.delimiterDataLine + formaters[l/2](data.cols[l+1].values[k]);
			}
			content += params.afterDataLine;
		}
		content += params.afterData;
	}
	else
	{
		var maxLength = 0;

		for(var j = 0; j < data.cols.length/2; j++){
			if(j!=0) { content += params.delimiterHead; }
			content += params.timeHead + params.delimiterHead + params.columnHeads[j];
			var log_length = Math.max(data.cols[j*2].log.length, data.cols[j*2+1].log.length)
			if (log_length > 0) {
				maxLength = Math.max(maxLength, log_length);
			}
			else {
				maxLength = Math.max(maxLength, data.cols[j*2].values.length);
			}
		}

		content += params.afterHead;

		for (var i = 0; i < maxLength ; i++){
			content += params.beforeDataLine;
			for (var k = 0; k < data.cols.length/2; k++){
				if(k!=0) { content += params.delimiterDataLine; }
				var log_length = Math.max(data.cols[k*2].log.length, data.cols[k*2+1].log.length);
				if (log_length > 0) {
					content += ((data.cols[k*2].log.length > i) ? data.cols[k*2].log[i] : "") + params.delimiterDataLine + ((data.cols[k*2+1].log.length > i) ? data.cols[k*2+1].log[i] : "");
				}
				else if(data.cols[k*2].values.length <= i){
					content += params.delimiterDataLine;
				}
				else
				{
					content += time_formater(data.cols[k*2].values[i]) + params.delimiterDataLine + formaters[k](data.cols[k*2+1].values[i]);
				}
			}
			content += params.afterDataLine;
		}
		content += params.afterData;
	}

	content += params.afterAll;
	return content;
};

VLH.formatCSV = function(data,params){
	VLH.checkOption(params,"afterHead","\n");
	VLH.checkOption(params,"delimiterHead",";");
	VLH.checkOption(params,"afterDataLine","\n");
	VLH.checkOption(params,"delimiterDataLine",";");
	return VLH.formatUniversal(data,params);
};

VLH.formatHTML = function(data,params){
	params.beforeAll = "<table><thead><tr><th>";
	params.delimiterHead = "</th><th>";
	params.afterHead = "</th></tr></thead><tbody>";
	params.beforeDataLine = "<tr><td>";
	params.delimiterDataLine = "</td><td>";
	params.afterDataLine = "</td></tr>";
	params.afterAll = "</tbody></table>";
	return VLH.formatUniversal(data,params);
};

VLH.formatDependentUniversal = function(data,params){
	VLH.checkOption(params,"beforeAll","");
	VLH.checkOption(params,"afterAll","");
	VLH.checkOption(params,"afterHead","");
	VLH.checkOption(params,"delimiterHead","");
	VLH.checkOption(params,"betweenHeadLines","");
	VLH.checkOption(params,"afterData","");
	VLH.checkOption(params,"beforeDataLine","");
	VLH.checkOption(params,"afterDataLine","");
	VLH.checkOption(params,"delimiterDataLine","");

	VLH.checkOption(params,"columnHeads","Series");
	VLH.checkOption(params,"valueFormats","0.000");
	VLH.checkOption(params,"series", []);

	if ((! (params.series instanceof Array)) || (params.series.length == 0)) {
		alert("Experiment data error : series parameter must be non-empty array.");
	}

	var column_count = params.series.length;

	if ((! (params.columnHeads instanceof Array)) || (params.columnHeads.length != column_count)) {
		alert("Experiment data error : columnHeads parameter must have the same number of elements as series array.");
	}

	if ((! (params.valueFormats instanceof Array)) || (params.valueFormats.length != column_count)) {
		alert("Experiment data error : valueFormats parameter must have the same number of elements as series array.");
	}

	// Determine how many rows are defined in the header.

	var head_rows = 1;
	for(var j = 0; j < column_count; j++) {
		var rows_in_column = 1;
		if (params.columnHeads[j] instanceof Array) {
			rows_in_column = params.columnHeads[j].length;
		}
		head_rows = Math.max(rows_in_column, head_rows);
	}

	// Extract the columns requested by the user.

	var columns = new Array();
	for (var i = 0; i < column_count; ++i) {
		var name = params.series[i];
		if (! (name in data.series)) {
			alert("Experiment data error : Series " + name + " is not provided by the data source.")
		}
		columns.push(data.series[name]);
	}

	// Prepare value formaters.

	var formaters = new Array();
	for (var i = 0; i < column_count; ++i) {
		formaters.push(VLH.getValueFormater(params.valueFormats[i]));
	}

	// TODO: Handle errors in the data.

	var was_error = false;
	for (var i = 0; i < column_count; i++){
		was_error = was_error || (columns[i].log.length > 0);
	}

	// Generate entries only for the portion where there is value in each column

	var min_value_count = columns.map(function(x) { return x.values.length; }).reduce(function (a, b) { return Math.min(a, b); });

	// Column header(s)

	var content = params.beforeAll;
	for (var i = 0; i < head_rows; ++i) {
		if (i > 0) {
			content += params.betweenHeadLines;
		}
		content += VLH.getArrayEntryWithDefault(params.columnHeads[0], i, "");
		for(var j = 1; j < column_count; j++) {
			content += params.delimiterHead + VLH.getArrayEntryWithDefault(params.columnHeads[j], i, "");
		}
	}
	content += params.afterHead;

	// Data

	for (var k = 0; k < min_value_count;k++) {
		content += params.beforeDataLine;
		content += formaters[0](columns[0].values[k]);
		for (var l = 1; l < column_count; ++l) {
			content += params.delimiterDataLine + formaters[l](columns[l].values[k]);
		}
		content += params.afterDataLine;
	}
	content += params.afterData;

	content += params.afterAll;
	return content;
};

VLH.checkOption = function(options,field,defaultValue){
	if (! (field in options)){
		options[field] = defaultValue;
	}
};

VLH.backup = function(object){
	object._initialStyles = $.extend(true,{},object.element[0].style);
};

VLH.undefine = function(object,property,undefineFunction){
	if (object[property] != undefined){
		undefineFunction(object[property]);
		object[property] = undefined;
	}
};


VLH.checkSynchronousData = function(data)
{
	if (data.series.length<=2) { return true; }

	var dataLength = data.cols[0].values.length;
	for (var k = 1; k < data.cols.length;k++)
	{
		if (data.cols[k].values.length != dataLength) { return false; }
	}

	for(var i = 0; i < dataLength; i++){
		var time = data.cols[0].values[i];
		for (var j = 2; j < data.cols.length; j += 2)
		{
			if(data.cols[j].values[i]!=time){ return false; }
		}
	}
	return true;
};

VLH.L = function(string){
	if (string in VLL){
		return VLL[string];
	} else {
		return "**" + string + "**";
	}
};
