/**
 * iSES Remote Lab SDK (c) by J.Dvorak, P.Kuriscak, F.Lustig
 *
 * This file is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 *
 * Copy of license along with attribution guidelines is distributed together with this work in LICENSE.txt file.
 * License text is also available online on <http://creativecommons.org/licenses/by-nc-sa/4.0/>.
 */

/* DATA PROCESSORS */

var VLP = {};

/**
 * Adds two numerical data columns together (value+value, disregarding time information).
 * @param params
 * {
 * 		firstName  - name of first series to add
 *   	secondName - name of second series to add to
 *   	resultName - name of resulting series
 *      onlyOverlapping - (optional, default false) sum only if both summands are present.
 * }
 */
VLP.Add = function(params,data,memory){
	VLH.checkOption(params,"onlyOverlapping",false);

	if (!(("firstName" in params) && ("secondName" in params))){
		alert("Add error: firstName and secondName must be specified!");
		return;
	}

	VLH.checkOption(params,"resultName",params.firstName+"+"+params.secondName);
	var result = VLH.InitData([params.resultName]);

	var firstSeries = data.series[params.firstName];
	var secondSeries = data.series[params.secondName];

	var firstLength = firstSeries.values.length;
	var secondLength = secondSeries.values.length;

	var overlappingLength = Math.min(firstLength,secondLength);
	var resultLength = overlappingLength;
	var longerSeries = firstSeries;
	if(!params.onlyOverlapping){
		if (firstLength < secondLength){
			longerSeries = secondSeries;
			resultLength = secondLength;
		} else {
			resultLength = firstLength;
		}
	}

	result.cols[0].values = [];

	for (var i = 0; i < overlappingLength;i++){
		result.cols[0].values.push(firstSeries.values[i]+secondSeries.values[i]);
	}

	for (var i = overlappingLength; i< resultLength ;i++){
		result.cols[0].values.push(longerSeries.values[i]);
	}

	result.cols[0].log = firstSeries.log.concat(secondSeries.log);

return result;
};

/**
 * Duplicates data series to another one with custom name
 * @param params
	{
		sourceName - name of the series to duplicate
		targetName - name of the duplicated series
	}
 */
VLP.Duplicate = function(params,data,memory){

	if (!(("sourceName" in params) && ("targetName" in params))){
			alert("Duplicate error: sourceName and targetName must be specified!");
			return;
		}

	var result = VLH.InitData([params.targetName]);
	result.cols[0].values = data.series[params.sourceName].values;
	result.cols[0].log = data.series[params.sourceName].log;
	return result;
};

/**
 * Drops samples which have the threshold-equal values in all series
 *
 * If the series contain different number of values, the result will have number
 * of values corresponding to length of the shortest series.
 *
 * @param params
 *  {
 *      sourceName - name of the source value, might be array of names
 *      targetName - name of the output value, if sourceName is
 *                   array, this must be array with the same length
 *      threshold - name of threshold values, if sourceName is
 *                   array, this must be array with the same length.
 *                   If not provided, 0 threshold will be used
 */
VLP.DropDuplicates = function(params, data, memory) {

	// Check parameters.

	if (!(("sourceName" in params) && ("targetName" in params))) {
		alert("DropDuplicates error: sourceName and targetName must be specified!");
		return;
	}
	if ((params.sourceName instanceof Array) != (params.targetName instanceof Array)) {
		alert("DropDuplicates error: sourceName and targetName must be both arrays or both scalars!");
		return;
	}
	if ((params.sourceName instanceof Array) && (params.sourceName.length != params.targetName.length)) {
		alert("DropDuplicates error: sourceName and targetName must be have the same length!");
		return;
	}
	if ("threshold" in params) {
		if ((params.sourceName instanceof Array) != (params.threshold instanceof Array)) {
			alert("DropDuplicates error: sourceName and threshold must be both arrays or both scalars!");
			return;
		}
		if ((params.sourceName instanceof Array) && (params.sourceName.length != params.threshold.length)) {
			alert("DropDuplicates error: sourceName and threshold must be have the same length!");
			return;
		}
	}

	var result;

	// Single column

	if (! (params.sourceName instanceof Array)) {

		var values = data.series[params.sourceName].values;
		result = VLH.InitData([params.targetName]);
		var out_values = result.cols[0].values;

		var value_count = values.length;
		var threshold = ("threshold" in params) ? params.threshold : 0;

		if (value_count > 0) {
			out_values.push(values[0]);
			var last_output_index = 0;

			for (var i = 1; i < value_count; ++i) {
				if (Math.abs(values[last_output_index] - values[i]) > threshold) {
					out_values.push(values[i]);
					last_output_index = i;
				}
			}
		}

		result.cols[0].log = data.series[params.sourceName].log;
	}

	// Multiple columns.

	else {
		var column_values = params.sourceName.map(function(name) { return data.series[name].values});
		result = VLH.InitData(params.targetName);

		var out_columns = result.cols.map(function(col) { return col.values; });

		var column_count = column_values.length;
		var value_count = column_values.map(function(x) { return x.length; }).reduce(function (a, b) { return Math.min(a, b); });
		var thresholds;
		if ("threshold" in params) {
			thresholds = params.threshold;
		}
		else {
			thresholds = [];
			for (var i = 0; i < column_count; ++i) {
				thresholds.push(0);
			}
		}

		if (value_count > 0) {

			for (var j = 0; j < column_count; ++j) {
				out_columns[j].push(column_values[j][0]);
			}
			var last_output_index = 0;

			for (var i = 1; i < value_count; ++i) {
				var match = 1;
				for (var j = 0; j < column_count; ++j) {
					if (Math.abs(column_values[j][last_output_index] - column_values[j][i]) > thresholds[j]) {
						match = 0;
						break;
					}
				}
				if (! match) {
					last_output_index = i;
					for (var j = 0; j < column_count; ++j) {
						out_columns[j].push(column_values[j][i]);
					}
				}
			}
		}

		for (var j = 0; j < column_count; ++j) {
			result.cols[j].log = data.series[params.sourceName[j]].log;
		}
	}

	return result;
};

/**
 * Keeps only sample pairs for which the value in the reference series is within
 * specified range.
 *
 * If the series contain different number of values, the result will have number
 * of values corresponding to length of the shortest series.
 *
 * @param params
 *  {
 *      sourceRefName - name of source reference (e.g. "series.time")
 *      sourceName - name of source pair value(s) (e.g. "series"), might be array of names
 *      targetRefName - name of output reference (e.g. "out_series.time")
 *      targetName - name of output pair value(s)  (e.g. "out_series"), if sourceName is
 *                   array, this must be array with the same length
 *      minValue - lower end of the range, if not provided there is no minimum
 *      minValueInclusive - does the minimal value belong to the range
 *      maxValue - upper end of the range, if not provided there is no maximum
 *      maxValueInclusive - does the maximal value belong to the range
 *  }
 */
VLP.ValueFilter = function(params, data,memory) {
	if (!(("sourceRefName" in params) && ("sourceName" in params) && ("targetRefName" in params) && ("targetName" in params))) {
		alert("ValueFilter error: sourceRefName, sourceName, targetRefName and targetName must be specified!");
		return;
	}
	if ((params.sourceName instanceof Array) != (params.targetName instanceof Array)) {
		alert("ValueFilter error: sourceName and targetName must be both arrays or both scalars!");
		return;
	}
	if ((params.sourceName instanceof Array) && (params.sourceName.length != params.targetName.length)) {
		alert("ValueFilter error: sourceName and targetName must be have the same length!");
		return;
	}

	// Create proper filter.

	var filter;
	if (params.minValue === undefined) {
		if (params.maxValue === undefined) {
			filter = function(value) { return true; }
		}
		else if (params.maxValueInclusive) {
			filter = function(value) { return (value <= params.maxValue); }
		}
		else {
			filter = function(value) { return (value < params.maxValue); }
		}
	}
	else if (params.maxValue === undefined) {
		if (params.minValueInclusive) {
			filter = function(value) { return (params.minValue <= value); }
		}
		else {
			filter = function(value) { return (params.minValue < value); }
		}
	}
	else if (params.minValueInclusive && params.maxValueInclusive) {
		filter = function(value) { return (params.minValue <= value) && (value <= params.maxValue); }
	}
	else if (params.minValueInclusive && (! params.maxValueInclusive)) {
		filter = function(value) { return (params.minValue <= value) && (value < params.maxValue); }
	}
	else if ((! params.minValueInclusive) && params.maxValueInclusive) {
		filter = function(value) { return (params.minValue < value) && (value <= params.maxValue); }
	}
	else {
		filter = function(value) { return (params.minValue < value) && (value < params.maxValue); }
	}

	var result;

	var refs = data.series[params.sourceRefName].values;

	// Single value per reference.

	if (! (params.sourceName instanceof Array)) {

		var values = data.series[params.sourceName].values;
		result = VLH.InitData([params.targetRefName, params.targetName]);

		var out_refs = result.cols[0].values;
		var out_values = result.cols[1].values;

		var value_count = Math.min(refs.length, values.length);

		for (var i = 0; i < value_count; ++i) {
			var ref = refs[i];
			if (filter(ref)) {
				out_refs.push(ref);
				out_values.push(values[i]);
			}
		}

		result.cols[0].log = data.series[params.sourceRefName].log;
		result.cols[1].log = data.series[params.sourceName].log;
	}

	// Multiple values per reference.

	else {
		var column_values = params.sourceName.map(function(name) { return data.series[name].values});
		result = VLH.InitData([params.targetRefName].concat(params.targetName));

		var out_refs = result.cols[0].values;
		var out_columns = result.cols.map(function(col) { return col.values; }).slice(1);

		var column_count = column_values.length;
		var value_count = column_values.map(function(x) { return x.length; }).reduce(function (a, b) { return Math.min(a, b); }, refs.length);

		for (var i = 0; i < value_count; ++i) {
			var ref = refs[i];
			if (filter(ref)) {
				out_refs.push(ref);
				for (var j = 0; j < column_count; ++j) {
					out_columns[j].push(column_values[j][i]);
				}
			}
		}

		result.cols[0].log = data.series[params.sourceRefName].log;
		for (var j = 0; j < column_count; ++j) {
			result.cols[1 + j].log = data.series[params.sourceName[j]].log;
		}
	}

	return result;
};

/**
 * Calculates average value and RMSQ of continuous sequences of samples
 * which have the same value in control series.
 *
 * @param params
	{
		countLimit - maximal number of values averaged into single output, if zero or not provided, the count is not limited
		controlName - name of serie with control value, might be array with more than one series
		controlTargetName - name of series to which to store values of the control series for individual samples, if empty strings or not provided, output is ignored
		controlThreshold - maximal delta between first and last averaged sample, if controlName is series is array, this must be also array. if not provided, assumed to be 0
		countName - name of serie to which to store number of samples used to calculate individual averages, if empty or not provided, counts are not stored
		series [{
			sourceName - name of serie whose value should be averaged
			targetName - name of resulting serie to store the value to
			rmsqName - name of serie to store the RMSQ to, if empty or not provided, the RMSQ is not stored
		}]
	}
 */
VLP.ControlledAverage = function(params,data,memory){

	if (!(("controlName" in params) && ("series" in params))){
		alert("ControlledAverage error: controlName and series must be specified!");
		return;
	}
	if (("controlTargetName" in params) && (VLH.getArrayLength(params.controlName) != VLH.getArrayLength(params.controlTargetName))) {
		alert("ControlledAverage error: if controlTargetName is provided, it must have the same number of elements as controlName!");
		return;
	}
	if (("controlThreshold" in params) && (VLH.getArrayLength(params.controlName) != VLH.getArrayLength(params.controlThreshold))) {
		alert("ControlledAverage error: if controlThreshold is provided, it must have the same number of elements as controlName!");
		return;
	}

	var outputSeries = [];

	// Prepare the reference series.

	var referenceValues = [];
	var controlSeriesRef = [];
	var controlSeriesOutputRefs = [];

	for (var i = 0; i < VLH.getArrayLength(params.controlName); ++i) {
		controlSeriesRef.push(VLH.getSeries(data, VLH.getArrayEntryWithDefault(params.controlName, i, "")));
		referenceValues.push(undefined);

		if ("controlTargetName" in params) {
			var controlTargetSeries = VLH.getArrayEntryWithDefault(params.controlTargetName, i, "");
			if (controlTargetSeries != "") {
				outputSeries.push(controlTargetSeries);
				controlSeriesOutputRefs.push(controlTargetSeries);
			}
			else {
				controlSeriesOutputRefs.push(null);
			}
		}
		else {
			controlSeriesOutputRefs.push(null);
		}
	}

	var sampleCount = controlSeriesRef.map(function(x) { return x.values.length; }).reduce(function (a, b) { return Math.min(a, b); });

	// Determine limits.

	var countLimit = sampleCount;
	if (("countLimit" in params) && (params.countLimit > 0)) {
		countLimit = params.countLimit;
	}

	// Determine thresholds.

	var thresholds = [];
	if ("controlThreshold" in params) {
		for (var i = 0; i < referenceValues.length; ++i) {
			thresholds.push(VLH.getArrayEntryWithDefault(params.controlThreshold, i, 0));
		}
	}
	else {
		for (var i = 0; i < referenceValues.length; ++i) {
			thresholds.push(0);
		}
	}

	// Get reference to all relevant source series.

	var countRef = undefined;
	if (("countName" in params) && (params.countName != "")) {
		outputSeries.push(params.countName);
	}
	else {
		countRef = null;
	}

	var seriesRefs = [];
	for (var i = 0; i < params.series.length; ++i) {
		var seriesRef = {};
		seriesRef.source = VLH.getSeries(data, params.series[i].sourceName);
		outputSeries.push(params.series[i].targetName);
		if (("rmsqName" in params.series[i]) && (params.series[i].rmsqName != "")) {
			outputSeries.push(params.series[i].rmsqName);
		}
		else {
			seriesRef.rmsq = null;
		}
		seriesRefs.push(seriesRef);

		sampleCount = Math.min(sampleCount, seriesRef.source.values.length);
	}

	// Initialize the output data.

	var result = VLH.InitData(outputSeries);

	// Update references to point to newly created series.

	if (countRef === undefined) {
		countRef = VLH.getSeries(result, params.countName);
	}
	for (var i = 0; i < params.series.length; ++i) {
		seriesRefs[i].target = VLH.getSeries(result, params.series[i].targetName);
		seriesRefs[i].target.log = seriesRef.source.log;
		if (! ("rmsq" in seriesRefs[i])) {
			seriesRefs[i].rmsq = VLH.getSeries(result, params.series[i].rmsqName);
			seriesRefs[i].rmsq.log = seriesRef.source.log;
		}
	}
	for (var i = 0; i < controlSeriesOutputRefs.length; ++i) {
		if (controlSeriesOutputRefs[i] != null) {
			controlSeriesOutputRefs[i] = VLH.getSeries(result, controlSeriesOutputRefs[i]);
		}
	}

	// Process all blocks.

	var sameRefSampleCount = 0;
	for (var i = 0; i < sampleCount; ++i) {

		// Is the value of all reference series the same as the previous output?

		var reference_match = true;
		for (var j = 0; j < controlSeriesRef.length; ++j) {
			if ((referenceValues[j] == undefined) || (Math.abs(referenceValues[j] - controlSeriesRef[j].values[i]) > thresholds[j])) {
				reference_match = false;
				break;
			}
		}

		// Stop the averaging if the number of samples is too big.

		if (sameRefSampleCount >= countLimit) {
			reference_match = false;
		}

		if (reference_match) {
			++sameRefSampleCount;
			continue;
		}
		if (sameRefSampleCount > 0) {
			if (countRef != null) {
				countRef.values.push(sameRefSampleCount);
			}
			for (var s = 0; s < seriesRefs.length; ++s) {
				VLH.averageRange(seriesRefs[s].source.values, i - sameRefSampleCount, sameRefSampleCount, seriesRefs[s].target.values, ((seriesRefs[s].rmsq != null) ? seriesRefs[s].rmsq.values : null))
			}
		}

		sameRefSampleCount = 1;
		for (var j = 0; j < controlSeriesRef.length; ++j) {
			referenceValues[j] = controlSeriesRef[j].values[i];

			// Output values of the reference series if required. The
			// averaged values will be initialized when the block ends.

			if (controlSeriesOutputRefs[j] != null) {
				controlSeriesOutputRefs[j].values.push(referenceValues[j]);
			}
		}
	}

	// Handle the final iteration.

	if (sameRefSampleCount > 0) {
		if (countRef != null) {
			countRef.values.push(sameRefSampleCount);
		}
		for (var s = 0; s < seriesRefs.length; ++s) {
			VLH.averageRange(seriesRefs[s].source.values, sampleCount - sameRefSampleCount, sameRefSampleCount, seriesRefs[s].target.values, ((seriesRefs[s].rmsq != null) ? seriesRefs[s].rmsq.values : null))
		}
	}
	return result;
};

/**
 * Keeps only those values from all series which have timewise close
 * counterpart in all other series.
 *
 * @param params
	{
		sourceNames - name of the series
		sourceTimeNames - name of source serie with times corresponding to sourceNames (e.g. "series_x.time"), defaults to "sourceName.time"
		targetNames - name of series with the aligned samples
		targetTimeNames - name of series with with the times corresponding to targetNames (e.g. "series_x.time"), defaults to "sourceName.time"
		pairTimeThreshold - maximal time distance between time values for them to be considered a close match
	}
 */
VLP.AlignSamples = function(params, data, memory) {

	// Check parameters.

	if (!(("sourceNames" in params) && ("targetNames" in params))) {
		alert("AlignSamples error: sourceNames and targetNames must be specified!");
		return;
	}
	if (params.sourceNames.length != params.targetNames.length) {
		alert("AlignSamples error: sourceNames and targetNames must have the same length!");
		return;
	}
	if (("sourceTimeNames" in params) && (params.sourceNames.length != params.sourceTimeNames.length)) {
		alert("AlignSamples error: sourceNames and sourceTimeNames must have the same length!");
		return;
	}
	if (("targetTimeNames" in params) && (params.targetNames.length != params.targetTimeNames.length)) {
		alert("AlignSamples error: targetNames and targetTimeNames must have the same length!");
		return;
	}

	// Generate the default time names.

	var sourceTimeNames = [];
	if ("sourceTimeNames" in params) {
		sourceTimeNames = params.sourceTimeNames;
	}
	else {
		for (var i = 0; i < params.sourceNames.length; ++i) {
			sourceTimeNames.push(params.sourceNames[i] + ".time");
		}
	}

	var targetTimeNames = [];
	if ("targetTimeNames" in params) {
		targetTimeNames = params.targetTimeNames;
	}
	else {
		for (var i = 0; i < params.targetNames.length; ++i) {
			targetTimeNames.push(params.targetNames[i] + ".time");
		}
	}

	// Locate sources.

	var sources = [];
	var sourceTimes = [];
	for (var i = 0; i < params.sourceNames.length; ++i) {
		sources.push(VLH.getSeries(data, params.sourceNames[i]));
		sourceTimes.push(VLH.getSeries(data, sourceTimeNames[i]));
	}

	// Prepare outputs and link the logs.

	var targets = [];
	var targetTimes = [];

	var result = VLH.InitData(params.targetNames.concat(targetTimeNames));
	for (var i = 0; i < sources.length; ++i) {
		var target = VLH.getSeries(result, params.targetNames[i]);
		target.log = sources[i].log;
		targets.push(target);

		var targetTime = VLH.getSeries(result, targetTimeNames[i]);
		targetTime.log = sourceTimes[i].log;
		targetTimes.push(targetTime);
	}

	var pairThreshold = ("pairTimeThreshold" in params) ? params.pairTimeThreshold : 0.0;

	// Initialize state tracking. Note that if any input
	// series is empty, there can be no output.

	var currentIndices = [];
	for (var i = 0; i < sources.length; ++i) {
		currentIndices.push(0);
		if (sources[i].values.length == 0) {
			return result;
		}
	}

	for (var i = 0; i < sourceTimes[0].values.length; ++i) {

		// The first series servers as master for all the matching.

		var referenceTime = sourceTimes[0].values[i];
		var referenceMin = referenceTime - pairThreshold;
		var referenceMax = referenceTime + pairThreshold;

		// Try to align other series to points which are sufficiently close to
		// the reference.

		var seriesAligned = true;
		for (var j = 1; seriesAligned && (j < sourceTimes.length); ++j) {
			for (;;) {

				// If we did run out of samples in any series, we are done.

				if (currentIndices[j] >= sourceTimes[j].values.length) {
					return result;
				}

				// If the current sample in this series is too old, advance to next
				// one.

				if (sourceTimes[j].values[currentIndices[j]] < referenceMin) {
					++currentIndices[j];
					continue;
				}

				// If there are more samples and the next sample is closer to the
				// reference time than this one advance to the next one.

				if ((currentIndices[j] + 1) < sourceTimes[j].values.length) {
					var currentDistance = Math.abs(sourceTimes[j].values[currentIndices[j]] - referenceTime);
					var nextDistance = Math.abs(sourceTimes[j].values[currentIndices[j] + 1] - referenceTime);
					if (nextDistance < currentDistance) {
						++currentIndices[j];
						continue;
					}
				}

				// If the value we found is within the required range, we
				// are aligned.

				seriesAligned = sourceTimes[j].values[currentIndices[j]] < referenceMax;
				break;
			}
		}

		// If it was not possible to get aligment for this master time, try
		// next master sample.

		if (! seriesAligned) {
			continue;
		}

		// Generate output sample and skip corresponding samples in all series.

		targetTimes[0].values.push(referenceTime);
		targets[0].values.push(sources[0].values[i]);

		for (var j = 1; j < sourceTimes.length; ++j) {
			targetTimes[j].values.push(sourceTimes[j].values[currentIndices[j]]);
			targets[j].values.push(sources[j].values[currentIndices[j]]);
			++currentIndices[j];
		}
	}

	return result;

};


/**
 * Stores up to specified number of aligned samples in buffer adding samples with
 * newer times when they become available in input.
 *
 * @param params
	{
		sourceNames - names of the series, there must be 1:1 relation between samples in individual series
		              for example generated by VLP.AlignSamples
		sourceTimeName - name of source serie with times corresponding to individual samples, ".time" series corresponding to first
		                 source is used
		targetNames - names of series with the memory samples, must have same number of entries as sourceNames, if not provided equals to sourceNames
		targetTimeName - name of serie with times corresponding to individual memory samples, defaults to same as sourceTimeName
		historyTimeThreshold - minimal time distance between two consecutive values which are added to the memory, 0 represents minimal nonzero time delta
		timeRange - maximal time range between first and last entry in the history, older entries are dropped, 0 means no range filtering
		sampleLimit - maximal number of samples in history, older entries are dropped. 0 or not present - disabled
	}
 */
VLP.Memory = function(params,data,memory){

	// Check parameters.

	if (!("sourceNames" in params)) {
		alert("Memory error: sourceNames must be specified!");
		return;
	}
	if (params.sourceNames.length == 0) {
		alert("Memory error: sourceNames can not be empty!");
		return;
	}

	// Fill in the defaults.

	var sourceTimeName = VLH.getParamWithDefault(params, "sourceTimeName", params.sourceNames[0] + ".time");
	var targetNames = VLH.getParamWithDefault(params, "targetNames", params.sourceNames);
	var targetTimeName = VLH.getParamWithDefault(params, "targetTimeName", sourceTimeName);

	if (params.sourceNames.length != params.targetNames.length) {
		alert("Memory error: sourceNames and targetNames must have the same length!");
		return;
	}

	var seriesCount = params.sourceNames.length;

	// Initialize the memory structure when we are called
	// for the first time.

	if (! ("values" in memory)) {
		memory.timeValues = [];
		memory.values = [];
		for (var i = 0; i < seriesCount; ++i) {
			memory.values.push([]);
		}
		memory.lastTime = null;
	}

	// Locate the input series.

	var sourceValues = [];
	for (var i = 0; i < seriesCount; ++i) {
		sourceValues.push(VLH.getSeries(data, params.sourceNames[i]).values);
	}
	var sourceTimes = VLH.getSeries(data, sourceTimeName).values;

	// Determine number of samples common to all series.

	var sampleCount = Math.min(sourceTimes.length, sourceValues.map(function(x) { return x.length; }).reduce(function (a, b) { return Math.min(a, b); }));

	// Determine first sample which is newer than last
	// remembered time.

	var index = 0;

	if (memory.lastTime !== null) {
		for (; (index < sampleCount) && (sourceTimes[index] <= memory.lastTime); ++index) {}
	}

	var historyThreshold = VLH.getParamWithDefault(params, "historyTimeThreshold", 0.0);

	while (index < sampleCount) {

		// Ignore samples which are too close to the last output.

		var time = sourceTimes[index];
		if (memory.lastTime !== null) {
			if ((memory.lastTime + historyThreshold) >= time) {
				++index;
				continue;
			}
		}

		// Store the sample.

		memory.lastTime = time;
		memory.timeValues.push(time);

		for (var i = 0; i < seriesCount; ++i) {
			memory.values[i].push(sourceValues[i][index]);
		}
	}

	// Prune old values by time or by count.

	if (("timeRange" in params) && (params.timeRange > 0) && (memory.timeValues.length > 1)) {
		var minTime = memory.timeValues[memory.timeValues.length - 1] - params.timeRange;
		while (memory.timeValues[0] < minTime) {
			memory.timeValues.shift();
			for (var i = 0; i < seriesCount; ++i) {
				memory.values[i].shift();
			}
		}
	}
	if (("sampleLimit" in params) && (params.sampleLimit > 0) && (memory.timeValues.length > params.sampleLimit)) {
		while (memory.timeValues.length > params.sampleLimit) {
			memory.timeValues.shift();
			for (var i = 0; i < seriesCount; ++i) {
				memory.values[i].shift();
			}
		}
	}

	// Return reference to our memory as output columns.

	var result = VLH.InitData([targetTimeName].concat(targetNames));
	result.cols[0].values = memory.timeValues;
	for (var i = 0; i < seriesCount; ++i) {
		result.cols[i + 1].values = memory.values[i];
	}
	return result;
};

/**
 * Stores pairs of X and Y values which have distinct time
 *
 * @param params
	{
		sourceXName - name of source serie with X values (e.g. "series_x")
		sourceYName - name of source serie with Y values (e.g. "series_y")
		sourceXTimeName - name of source serie with times corresponding to sourceXName values (e.g. "series_x.time"), defaults to "sourceXName.time"
		sourceYTimeName - name of source serie with times corresponding to sourceYName values (e.g. "series_y.time"), defaults to "sourceYName.time"
		targetXName - name of target serie with X values
		targetYName - name of target serie with Y values
		pairTimeThreshold - maximal time distance between time values for them to be considered a matching pair
		historyTimeThreshold - minimal time distance between two consecutive values which are added to the memory, 0 represents minimal nonzero time delta
		timeRange - maximal time range between first and last entry in the history, older entries are dropped, 0 means no range filtering
		sampleLimit - maximal number of samples in history, older entries are dropped. 0 or not present - disabled
	}
 */
VLP.XYMemory = function(params,data,memory){

	// Check parameters.

	if (!(("sourceXName" in params) && ("sourceYName" in params))) {
		alert("XYMemory error: sourceXName and sourceYName must be specified!");
		return;
	}
	if (!(("targetXName" in params) && ("targetYName" in params))) {
		alert("XYMemory error: targetXName and targetYName must be specified!");
		return;
	}

	// Initialize the memory structure when we are called
	// for the first time.

	if (! ("xValues" in memory)) {
		memory.timeValues = [];
		memory.xValues = [];
		memory.yValues = [];
		memory.lastTime = null;
	}

	// Update memory

	var xSeries = VLH.getSeries(data, params.sourceXName);
	var ySeries = VLH.getSeries(data, params.sourceYName);
	var xSeriesTime = VLH.getSeries(data, ("sourceXTimeName" in params) ? params.sourceXTimeName : params.sourceXName + ".time");
	var ySeriesTime = VLH.getSeries(data, ("sourceYTimeName" in params) ? params.sourceYTimeName : params.sourceYName + ".time");

	var xLength = Math.min(xSeries.values.length, xSeriesTime.values.length);
	var yLength = Math.min(ySeries.values.length, ySeriesTime.values.length);

	// Determine first sample which is newer than last
	// remembered time.

	var xIndex = 0;
	var yIndex = 0

	if (memory.lastTime !== null) {
		for (; (xIndex < xLength) && (xSeriesTime.values[xIndex] <= memory.lastTime); ++xIndex) {}
		for (; (yIndex < yLength) && (ySeriesTime.values[yIndex] <= memory.lastTime); ++yIndex) {}
	}

	var pairThreshold = ("pairTimeThreshold" in params) ? params.pairTimeThreshold : 0.0;
	var historyThreshold = ("historyTimeThreshold" in params) ? params.historyTimeThreshold : 0.0;

	while ((xIndex < xLength) && (yIndex < yLength)) {

		// Look for values with sufficiently close times

		var xTime = xSeriesTime.values[xIndex];
		var yTime = ySeriesTime.values[yIndex];
		if ((xTime + pairThreshold) < yTime) {
			++xIndex;
			continue;
		}
		else if ((yTime + pairThreshold) < xTime) {
			++yIndex;
			continue;
		}

		// Ignore samples which are too close to the last output.

		var maxTime = Math.max(xTime, yTime);
		if (memory.lastTime !== null) {
			if ((memory.lastTime + historyThreshold) >= maxTime) {
				++xIndex;
				++yIndex;
				continue;
			}
		}

		// Store the sample.

		memory.lastTime = maxTime;
		memory.timeValues.push(maxTime);
		memory.xValues.push(xSeries.values[xIndex]);
		memory.yValues.push(ySeries.values[yIndex]);
	}

	// Prune old values by time or by count.

	if (("timeRange" in params) && (params.timeRange > 0) && (memory.timeValues.length > 1)) {
		var minTime = memory.timeValues[memory.timeValues.length - 1] - params.timeRange;
		while (memory.timeValues[0] < minTime) {
			memory.timeValues.shift();
			memory.xValues.shift();
			memory.yValues.shift();
		}
	}
	if (("sampleLimit" in params) && (params.sampleLimit > 0) && (memory.timeValues.length > params.sampleLimit)) {
		while (memory.timeValues.length > params.sampleLimit) {
			memory.timeValues.shift();
			memory.xValues.shift();
			memory.yValues.shift();
		}
	}

	// Return reference to our memory as output columns.

	var result = VLH.InitData([params.targetXName, params.targetYName]);
	result.cols[0].values = memory.xValues;
	result.cols[1].values = memory.yValues;
	return result;
};

/**
 * Looks up integer representation of source value inside table
 *
 * @param params
	{
		sourceName - name of source serie with the base values
		targetName - name of target serie to which to store the colors
		lookupTable - map of values, indexed by value of sourceName converted to integer
	}
 */
VLP.Lookup = function(params,data,memory){

	// Check parameters.

	if (!(("sourceName" in params) && ("targetName" in params) && ("lookupTable" in params))) {
		alert("Lookup error: sourceName, targetName and lookupTable must be specified!");
		return;
	}
	if (params.lookupTable.length == 0) {
		alert("Lookup error: lookupTable can not be empty!");
		return;
	}

	// Update memory

	var sourceSeries = VLH.getSeries(data, params.sourceName);

	// Initialize the output.

	var result = VLH.InitData([params.targetName]);
	result.cols[0].log = sourceSeries.log;
	result.cols[0].values = [];

	var input = sourceSeries.values;
	var output = result.cols[0].values;
	var valueCount =  sourceSeries.values.length;
	var maxIndex = params.lookupTable.length - 1;

	for (var i = 0; i < valueCount; ++i) {
		var value = input[i];
		var index = Math.min(Math.max(0, parseInt(value)), maxIndex);
		output.push(params.lookupTable[index]);
	}

	return result;
};

/**
 * Applies function to input of one or more numerical columns.
 *
 * If the columns have different length, only the common prefix is processed.
 *
 * @param params
 * {
 * 		sourceNames - name of the input series
 *   	targetName - name of the resulting series
 *    combiningFunction - function to evaluate for each input, the input values are provided in array
 * }
 */
VLP.Function = function(params,data,memory){

	// Check parameters.

	if (!("sourceNames" in params)) {
		alert("Function error: sourceNames must be specified!");
		return;
	}
	if (params.sourceNames.length == 0) {
		alert("Function error: sourceNames can not be empty!");
		return;
	}

	if (!("targetName" in params)) {
		alert("Function error: targetName must be specified!");
		return;
	}
	if (!("combiningFunction" in params)) {
		alert("Function error: combiningFunction must be specified!");
		return;
	}

	// Setup the output series.

	var result = VLH.InitData([params.targetName]);
	var resultValues = result.cols[0].values;

	var seriesCount = params.sourceNames.length;

	// Locate data of individual series

	var sourceValues = [];
	for (var i = 0; i < seriesCount; ++i) {
		var series = VLH.getSeries(data, params.sourceNames[i]);
		sourceValues.push(series.values);
	}

	// Combine logs.

	for (var i = 0; i < seriesCount; ++i) {
		result.cols[0].log = result.cols[0].log.concat(series.log);
	}

	// Determine number of samples common to all series.

	var sampleCount = sourceValues.map(function(x) { return x.length; }).reduce(function (a, b) { return Math.min(a, b); });

	// Array used to store values passed as parameters to the
	// function.

	var valueArray = []
	for (var i = 0; i < seriesCount; ++i) {
		valueArray.push(0.0);
	}

	// Evaluate the result.

	for (var i = 0; i < sampleCount;i++){
		for (var j = 0; j < seriesCount; ++j) {
			valueArray[j] = sourceValues[j][i];
		}
		resultValues.push(params.combiningFunction(valueArray));
	}
	return result;
};
